#include "synth.h"

Synth::Synth(double sampleRate){
    setSampleRate(48000.0, sampleRate);
}

#define frac(x) ((x) - ((long)x))
constexpr double MIN_VOLUME = 0.00001;

//Volume of voice as a function of sample rate independent frequency.
static constexpr double getAmplitude( double hz )
{
  if( hz < 20.0 )       return 0.0;
  if( hz < 440.0 )      return hz / 440.0;
  if( hz < 10000.0 )    return (10000.0 - hz) / 9560.0;
  return 0.0;
}

//Sanity checks: voices should become silent outside audible frequencies.
//Voice index code will loop infinitely if these fail.
static_assert( MIN_VOLUME > getAmplitude(10.0) );
static_assert( MIN_VOLUME > getAmplitude(20000.0));

void Synth::process(float* output, const uint32_t frames)
{
    double hz;

    //Render.
    for(uint32_t i = 0; i < frames; i++){

        //Set pitch.
        hz = hzFund *= hzShift;

        //Set timbre.
        ratio = ratioSlewRate * ratio + (1.0 - ratioSlewRate) * targetRatio;
        
        for(uint voice = 0; voice < NUM_VOICES; ++voice){
            //Get new phase.
            double phase = phases[(voice + idxFund) & VOICE_MASK];
            phase += hz * sampleInterval;
            phase = frac(phase);

            //Don't bother rendering anything over the Nyquist rate.
            if( hz > hzNyq ) break;
            output[i] += getAmplitude(hz) * (phase > 0.5); // * sin(2.0 * M_PI * phase);

            //Remember phase, move to higher overtone.
            phases[(voice + idxFund) & VOICE_MASK] = phase;
            hz *= ratio;
        }

        output[i] *= volume / static_cast<double>(NUM_VOICES);
    }

    //Make timbre cyclical.
    if( hzShift < 1.0 )
    {
        //Pitch low and decreasing, shift bottom voices to top.
        while( getAmplitude(hz) > MIN_VOLUME
            && getAmplitude(hzFund) < MIN_VOLUME)
        {
            ++idxFund;
            hz *= ratio;
            hzFund *= ratio;
        }
    }

    if( (hzShift > 1.0) || (targetRatio > ratio))
    {
        //Pitch high and increasing. Shift top voices to bottom.
        while( getAmplitude(hz) < MIN_VOLUME
            && getAmplitude(hzFund) > MIN_VOLUME)
        {
            --idxFund;
            hz /= ratio;
            hzFund /= ratio;
        }
    }
}

void Synth::setSampleRate(double oldRate, double newRate){
    sampleInterval = 1.0 / newRate;
    hzNyq = newRate / 2;
    ratioSlewRate = pow(ratioSlewRate, oldRate / newRate);
    hzShift = pow(hzShift, newRate / oldRate);
}

//Takes value from 0 to 1 representing frequency shift factor.
//0 : lower one octave per second
//1 : raise one octave per second
void Synth::setFrequencyShift(double in){
    hzShift = exp2(2.0 * (in - 0.5) * sampleInterval);
}

static constexpr double minRatio = 1.0;
static constexpr double maxRatio = 3.0;
//Slew to given ratio.
//0 : next voice is one fifth  above previous voice.
//1 : next voice is one octave above previous voice.
void Synth::setTargetRatio(double in){
    targetRatio = minRatio + in * (maxRatio - minRatio);
}

void Synth::setVolume(double in){
    volume = in;
}